package school_project.Parsers;

import school_project.Map;
import school_project.Piece;
import school_project.Utils.Bitwise;

import java.io.*;
import java.util.Arrays;

public class BinaryParser implements FileParser {

    /**
     * Check if the file is a binary file suited for parsing
     * @return true if it is a binary File
     */
    static boolean isBinaryFile(File file){
        // TODO: 3/18/23
        return false;
    }

    public Map getLevel(File file) throws IOException {
        Map ret;

        FileInputStream fileStream = new FileInputStream(file);

        byte[] level_data = ExtractLevelData(fileStream);

        ret = new Map(ExtractMapFromLevelData(level_data));

        ret.addPiece(ExtractPiecesFromLevelData(level_data));

        fileStream.close();
        return ret;
    }

    /**
     * Get Pieces out of the level data
     * @param levelData full data of the level without header and footer
     * @return array of Piece from level data
     */
    private static Piece[] ExtractPiecesFromLevelData(byte[] levelData) {
        byte map_width = levelData[0], map_height = levelData[1];
        byte piece_count = levelData[3 + map_width * map_height / 8 + (map_height * map_width % 8 != 0 ? 1 : 0)];
        Piece[] ret = new Piece[piece_count];
        byte[] pieces_data = Arrays.copyOfRange(levelData, 4 + map_width * map_height / 8 + (map_height * map_width % 8 != 0 ? 1 : 0), levelData.length-1);
        for (int piece_index = 0; piece_index < piece_count; piece_index++) {
            byte[] _piece_size = Bitwise.ByteToNible(pieces_data[piece_index]);
            byte _piece_width = _piece_size[0], _piece_height = _piece_size[1];
            byte[] _piece_data = Arrays.copyOfRange(pieces_data, piece_index + 1, 1 + _piece_width * _piece_height / 8 + (_piece_height * _piece_width % 8 != 0 ? 1 : 0));
            boolean[][] _piece_matrix = BuildMatrixFromBytes(_piece_width, _piece_height, _piece_data);
            ret[piece_index] = new Piece(_piece_matrix);
        }
        return ret;
    }

    /**
     * Get the Map Matrix out of the level data
     * @param level_data full data of the level without header and footer
     * @return boolean matrix of the map
     */
    private static boolean[][] ExtractMapFromLevelData(byte[] level_data){
        int map_width = level_data[0], map_height = level_data[1];
        byte[] map_data = Arrays.copyOfRange(level_data, 2, 2 + map_width * map_height / 8 + (map_height * map_width % 8 != 0 ? 1 : 0));
        return BuildMatrixFromBytes(map_width, map_height, map_data);
    }

    /**
     * Extract Level data from file content
     * @param fileStream file stream to read extract data from
     * @return Level data as an array of byte
     * @throws IOException Expected if we can't read the file
     */
    private static byte[] ExtractLevelData(FileInputStream fileStream) throws IOException {

        byte[] bytes = fileStream.readAllBytes();

        int start_position = 0, end_position = 0;
        for (int i = 0; i < bytes.length; i++) {
            if(bytes[i] == 83 && bytes[i+1] == 77 && bytes[i+2] == 83){ // SMS
                start_position = i+3;
                break;
            }
        }

        for (int i = start_position; i < bytes.length; i++) {
            if(bytes[i] == 83 && bytes[i+1] == 77 && bytes[i+2] == 69){ // SME
                end_position = i;
                break;
            }
        }

        return Arrays.copyOfRange(bytes, start_position, end_position);
    }

    /**
     * Build a boolean Matrix From a byte array
     * Each Byte is composed of 8 bit, each bit is 1 or 0
     *  if the bit is 0 then it's a false for this cell
     *  else it's true for this cell
     * @param matrix_width width of the matrix
     * @param matrix_height height of the matrix
     * @param matrix_data byte array of the data to export
     * @return boolean Matrix of the data decompiled
     */
    private static boolean[][] BuildMatrixFromBytes(int matrix_width, int matrix_height, byte[] matrix_data){
        boolean[][] ret = new boolean[matrix_height][matrix_width];

        // Transforming the bit from matrix_data's byte into boolean array for better manipulation
        boolean[] b_array = new boolean[matrix_height * matrix_width];
        int index = 0;
        for(byte b: matrix_data){
            for (int i = 0; i < 8; i++) { // because 8 bit in a byte
                b_array[index] = Bitwise.IsBitSetAt(b, i);
                index++;
            }
        }

        // Transforming b_array to a 2D matrix
        for (int x = 0; x < matrix_height; x++) {
            for (int y = 0; y < matrix_width; y++) {
                ret[x][y] = b_array[y + x * 8];
            }
        }
        return ret;
    }
}